﻿using Ceras;
using Ceras.Helpers;
using ChatObjects;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ChatClient
{
    public partial class ChatForm : Form
    {
        SetupForm setupForm;
		Client client;

		public ChatForm()
        {
            InitializeComponent();
        }

        private void settingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (setupForm == null)
            {
                setupForm = new SetupForm();
            }
			
            if (setupForm.ShowDialog() == DialogResult.OK)
			{
				connectToolStripMenuItem.Enabled = true;
				this.Text = "ChatClient - " + setupForm.edUserName.Text;
			}
        }

		public void setEnabledItems(bool enabled)
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action<bool>(setEnabledItems), new object[] { enabled });
				return;
			}
			edMessage.Enabled = enabled;
			connectToolStripMenuItem.Enabled = !enabled;
			disconnectToolStripMenuItem.Enabled = enabled;
			settingsToolStripMenuItem.Enabled = !enabled;
			btnSend.Enabled = enabled;
			if (enabled)
			{
				toolStripStatusLabel.Text = "Status: connected";
			} else
			{
				toolStripStatusLabel.Text = "Status: disconnected";
			}
		}
        private void connectToolStripMenuItem1_Click(object sender, EventArgs e)
        {
			if (setupForm != null)
			{
				client = new Client();
				client.Start(this, setupForm.edUserName.Text,setupForm.edHost.Text, Int32.Parse(setupForm.edPort.Text));
			}			
		}

		private void btnSend_Click(object sender, EventArgs e)
		{
			if (client != null)
			{
				client.Send(edMessage.Text);
			}
		}

		public void AddMessageToTextBox(string msg)
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action<string>(AddMessageToTextBox), new object[] { msg });
				return;
			}
			edMessage.Text = "";
			txtMessages.AppendText(msg);
			txtMessages.AppendText(Environment.NewLine);
		}

		public void SetUserList(List<string> list)
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action<List<string>>(SetUserList), new object[] { list });
				return;
			}
			lbUsers.Items.Clear();
			list.ForEach(delegate (string name)
			{
				lbUsers.Items.Add(name);
			});
		}

		public void cleanUserList()
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action(cleanUserList), new object[] { });
				return;
			}
			lbUsers.Items.Clear();
		}

		private void edMessage_KeyPress(object sender, KeyPressEventArgs e)
		{
			if (client != null && e.KeyChar == (char)13)
			{
				client.Send(edMessage.Text);
			}
		}

		private void edMessage_KeyUp(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Enter)
			{
				e.SuppressKeyPress = false;
				e.Handled = true;
			}
		}

		private void disconnectToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (client != null)
			{
				clientDisconnect();
				setEnabledItems(false);
				lbUsers.Items.Clear();
			}
		}

		public void ChatForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (client != null)
			{
				clientDisconnect();
			}
		}

		public void clientDisconnect()
		{
			try
			{
				ClientLogout();
				client.Disconnect();
				client = null;
			}
			catch (Exception ex)
			{

			}
		}

		private void ClientLogout()
		{
			Logout logout = new Logout();
			logout.userName = setupForm.edUserName.Text;
			client.Send(logout);
		}

		public void AddUserName(string userName)
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action<string>(AddUserName), new object[] { userName });
				return;
			}
			lbUsers.Items.Add(userName);
		}

		public void RemoveUserName(string userName)
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action<string>(RemoveUserName), new object[] { userName });
				return;
			}
			lbUsers.Items.Remove(userName);
		}

		public void ShowUserWarnDialog()
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action(ShowUserWarnDialog), new object[] { });
				return;
			}
			MessageBox.Show("Same username exists", "User exists", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		internal void ShowConnectWarnDialog()
		{
			if (InvokeRequired)
			{
				this.Invoke(new Action(ShowConnectWarnDialog), new object[] { });
				return;
			}
			lbUsers.Items.Clear();
			MessageBox.Show("Connection lost. Please connect again.", "Connection lost.", MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		private void exitToolStripMenuItem_Click_1(object sender, EventArgs e)
		{
			Close();
		}
	}

	class Client
	{
		TcpClient client;
		NetworkStream netStream;
		CerasSerializer sendCeras;
		CerasSerializer receiveCeras;
		bool disconnected = false;
		ChatForm chatForm;
		string userName;

		public void Start(ChatForm chatForm, string userName, string host, int port)
		{
			this.chatForm = chatForm;
			this.userName = userName;

			// Create network connection	
			client = new TcpClient();
			try
			{
				client.Connect(host, port);
				// And use a network stream, much more comfortable to use
				netStream = client.GetStream();


				// Now we need our serializer
				// !! Important:
				// !! The settings of the serializers for client and server must be the same!
				var config1 = new SerializerConfig();
				config1.Advanced.PersistTypeCache = false;

				var config2 = new SerializerConfig();
				config2.Advanced.PersistTypeCache = false;

				sendCeras = new CerasSerializer(config1);
				receiveCeras = new CerasSerializer(config2);

				// Start a thread that receives and reacts to messages from the server
				StartReceiving();
				Login login = new Login();
				login.userName = this.userName;
				Send(login);
				Send(new GetUserList());
				chatForm.setEnabledItems(true);
			}
			catch (Exception ex)
			{
				chatForm.ShowConnectWarnDialog();
			}
		}

		void StartReceiving()
		{
			Task.Run(async () =>
			{
				try
				{
					while (!disconnected)
					{
						// Read until we received the next message from the server
						var obj = await receiveCeras.ReadFromStream(netStream);
						HandleMessage(obj);
					}
				}
				catch (Exception e)
				{
					Console.WriteLine("Client error while receiving: " + e);
				}
			});
		}

		void HandleMessage(object obj)
		{
			if (obj is string)
			{
				chatForm.AddMessageToTextBox((string)obj);
			} else if (obj is Login login)
			{
				if (login.isConnected)
				{
					chatForm.AddUserName(login.userName);
				} else
				{
					chatForm.setEnabledItems(false);
					Disconnect();
					chatForm.ShowUserWarnDialog();
				}
				chatForm.setEnabledItems(login.isConnected);
			} else if (obj is Logout logout)
			{
				chatForm.RemoveUserName(logout.userName);
			} else if (obj is GetUserList userList)
			{
				chatForm.SetUserList(userList.list);
			} else if (obj is ServerShutdown)
			{
				chatForm.setEnabledItems(false);
				chatForm.cleanUserList();
				Disconnect();
				chatForm.ShowConnectWarnDialog();
			}
		}

		// A little helper function that sends any object to the server
		public void Send(object obj)
		{
			try 
			{ 				
				if (client != null && client.Connected)
				{
					sendCeras.WriteToStream(netStream, obj);
				}
			} catch (Exception ex)
			{
				chatForm.setEnabledItems(false);
				Disconnect();
			}
		}

		public void Disconnect()
		{
			disconnected = true;
			if (netStream != null)
			{
				netStream.Close();
				netStream.Dispose();
				netStream = null;
			}
			if (client != null)
			{
				if (client.Connected)
				{
					client.Close();
				}
				client.Dispose();
				client = null;
			}
		}
	}
}
